@@ -20,6 +20,8 @@ package nats
 import (
 	"crypto/sha256"
 	"encoding/base64"
+	"encoding/json"
+	"errors"
 	"fmt"
 	"io/ioutil"
 	"math/rand"
@@ -291,7 +293,7 @@ func TestJetStreamOrderedConsumer(t *testing.T) {
 	testSyncConsumer()
 }
 
-func TestJetStreamOrderedConsumerWithErrors(t *testing.T) {
+func TestJetStreamOrderedConsumerDeleteAssets(t *testing.T) {
 	s := RunBasicJetStreamServer()
 	defer shutdownJSServerAndRemoveStorage(t, s)
 
@@ -339,15 +341,9 @@ func TestJetStreamOrderedConsumerWithErrors(t *testing.T) {
 		}
 	}
 
-	type asset int
-	const (
-		deleteStream asset = iota
-		deleteConsumer
-	)
+	t.Run("remove stream, expect error", func(t *testing.T) {
+		createStream()
 
-	testSubError := func(a asset) {
-		t.Helper()
-		// Again here the IdleHeartbeat is not required, just overriding top shorten test time.
 		sub, err := js.SubscribeSync("a", OrderedConsumer(), IdleHeartbeat(200*time.Millisecond))
 		if err != nil {
 			t.Fatalf("Unexpected error: %v", err)
@@ -357,37 +353,71 @@ func TestJetStreamOrderedConsumerWithErrors(t *testing.T) {
 		// Since we are sync we will be paused here due to flow control.
 		time.Sleep(100 * time.Millisecond)
 		// Now delete the asset and make sure we get an error.
-		switch a {
-		case deleteStream:
-			if err := js.DeleteStream("OBJECT"); err != nil {
-				t.Fatalf("Unexpected error: %v", err)
-			}
-		case deleteConsumer:
-			// We need to grab our consumer name.
-			ci, err := sub.ConsumerInfo()
-			if err != nil {
-				t.Fatalf("Unexpected error: %v", err)
-			}
-			if err := js.DeleteConsumer("OBJECT", ci.Name); err != nil {
-				t.Fatalf("Unexpected error: %v", err)
-			}
+		if err := js.DeleteStream("OBJECT"); err != nil {
+			t.Fatalf("Unexpected error: %v", err)
 		}
 		// Make sure we get an error.
 		select {
 		case err := <-errCh:
-			if err != ErrConsumerNotActive {
-				t.Fatalf("Got wrong error, wanted %v, got %v", ErrConsumerNotActive, err)
+			if !errors.Is(err, ErrStreamNotFound) {
+				t.Fatalf("Got wrong error, wanted %v, got %v", ErrStreamNotFound, err)
 			}
 		case <-time.After(time.Second):
 			t.Fatalf("Did not receive err message as expected")
 		}
-	}
+	})
 
-	createStream()
-	testSubError(deleteStream)
+	t.Run("remove consumer, expect it to be recreated", func(t *testing.T) {
+		createStream()
 
-	createStream()
-	testSubError(deleteConsumer)
+		createConsSub, err := nc.SubscribeSync("$JS.API.CONSUMER.CREATE.OBJECT")
+		if err != nil {
+			t.Fatalf("Unexpected error: %v", err)
+		}
+		defer createConsSub.Unsubscribe()
+		// Again here the IdleHeartbeat is not required, just overriding top shorten test time.
+		sub, err := js.SubscribeSync("a", OrderedConsumer(), IdleHeartbeat(200*time.Millisecond))
+		if err != nil {
+			t.Fatalf("Unexpected error: %v", err)
+		}
+		defer sub.Unsubscribe()
+
+		createConsMsg, err := createConsSub.NextMsg(time.Second)
+		if err != nil {
+			t.Fatalf("Unexpected error: %v", err)
+		}
+		if !strings.Contains(string(createConsMsg.Data), `"stream_name":"OBJECT"`) {
+			t.Fatalf("Invalid message on create consumer subject: %q", string(createConsMsg.Data))
+		}
+
+		time.Sleep(100 * time.Millisecond)
+		ci, err := sub.ConsumerInfo()
+		if err != nil {
+			t.Fatalf("Unexpected error: %v", err)
+		}
+		consName := ci.Name
+
+		if err := js.DeleteConsumer("OBJECT", consName); err != nil {
+			t.Fatalf("Unexpected error: %v", err)
+		}
+		createConsMsg, err = createConsSub.NextMsg(time.Second)
+		if err != nil {
+			t.Fatalf("Unexpected error: %v", err)
+		}
+		if !strings.Contains(string(createConsMsg.Data), `"stream_name":"OBJECT"`) {
+			t.Fatalf("Invalid message on create consumer subject: %q", string(createConsMsg.Data))
+		}
+
+		time.Sleep(100 * time.Millisecond)
+		ci, err = sub.ConsumerInfo()
+		if err != nil {
+			t.Fatalf("Unexpected error: %v", err)
+		}
+		newConsName := ci.Name
+		if consName == newConsName {
+			t.Fatalf("Consumer should be recreated, but consumer name is the same")
+		}
+	})
 }
 
 func TestJetStreamOrderedConsumerWithAutoUnsub(t *testing.T) {
@@ -843,7 +873,7 @@ func TestJetStreamFlowControlStalled(t *testing.T) {
 	}
 
 	// Publish bunch of messages.
-	payload := make([]byte, 1024)
+	payload := make([]byte, 100*1024)
 	for i := 0; i < 250; i++ {
 		nc.Publish("a", payload)
 	}
@@ -939,3 +969,303 @@ func TestJetStreamExpiredPullRequests(t *testing.T) {
 		}
 	}
 }
+
+func TestJetStreamSyncSubscribeWithMaxAckPending(t *testing.T) {
+	opts := natsserver.DefaultTestOptions
+	opts.Port = -1
+	opts.JetStream = true
+	opts.JetStreamLimits.MaxAckPending = 123
+	s := natsserver.RunServer(&opts)
+	defer shutdownJSServerAndRemoveStorage(t, s)
+
+	nc, js := jsClient(t, s)
+	defer nc.Close()
+
+	if _, err := js.AddStream(&StreamConfig{Name: "MAX_ACK_PENDING", Subjects: []string{"foo"}}); err != nil {
+		t.Fatalf("Error adding stream: %v", err)
+	}
+
+	// By default, the sync subscription will be created with a MaxAckPending equal
+	// to the internal sync queue len, which is 64K. So that should error out
+	// and make sure we get the actual limit
+
+	checkSub := func(pull bool) {
+		var sub *Subscription
+		var err error
+		if pull {
+			_, err = js.PullSubscribe("foo", "bar")
+		} else {
+			_, err = js.SubscribeSync("foo")
+		}
+		if err == nil || !strings.Contains(err.Error(), "system limit of 123") {
+			t.Fatalf("Unexpected error: %v", err)
+		}
+
+		// But it should work if we use MaxAckPending() with lower value
+		if pull {
+			sub, err = js.PullSubscribe("foo", "bar", MaxAckPending(64))
+		} else {
+			sub, err = js.SubscribeSync("foo", MaxAckPending(64))
+		}
+		if err != nil {
+			t.Fatalf("Unexpected error: %v", err)
+		}
+		sub.Unsubscribe()
+	}
+	checkSub(false)
+	checkSub(true)
+}
+
+func TestJetStreamClusterPlacement(t *testing.T) {
+	// There used to be a test here that would not work because it would require
+	// all servers in the cluster to know about each other tags. So we will simply
+	// verify that if a stream is configured with placement and tags, the proper
+	// "stream create" request is sent.
+	s := RunBasicJetStreamServer()
+	defer shutdownJSServerAndRemoveStorage(t, s)
+
+	nc, js := jsClient(t, s)
+	defer nc.Close()
+
+	sub, err := nc.SubscribeSync(fmt.Sprintf("$JS.API."+apiStreamCreateT, "TEST"))
+	if err != nil {
+		t.Fatalf("Error on sub: %v", err)
+	}
+	js.AddStream(&StreamConfig{
+		Name: "TEST",
+		Placement: &Placement{
+			Tags: []string{"my_tag"},
+		},
+	})
+	msg, err := sub.NextMsg(time.Second)
+	if err != nil {
+		t.Fatalf("Error getting stream create request: %v", err)
+	}
+	var req StreamConfig
+	if err := json.Unmarshal(msg.Data, &req); err != nil {
+		t.Fatalf("Unmarshal error: %v", err)
+	}
+	if req.Placement == nil {
+		t.Fatal("Expected placement, did not get it")
+	}
+	if n := len(req.Placement.Tags); n != 1 {
+		t.Fatalf("Expected 1 tag, got %v", n)
+	}
+	if v := req.Placement.Tags[0]; v != "my_tag" {
+		t.Fatalf("Unexpected tag: %q", v)
+	}
+}
+
+func TestJetStreamConvertDirectMsgResponseToMsg(t *testing.T) {
+	// This test checks the conversion of a "direct get message" response
+	// to a JS message based on the content of specific NATS headers.
+	// It is very specific to the order headers retrieval is made in
+	// convertDirectGetMsgResponseToMsg(), so it may need adjustment
+	// if changes are made there.
+
+	msg := NewMsg("inbox")
+
+	check := func(errTxt string) {
+		t.Helper()
+		m, err := convertDirectGetMsgResponseToMsg("test", msg)
+		if err == nil || !strings.Contains(err.Error(), errTxt) {
+			t.Fatalf("Expected error contain %q, got %v", errTxt, err)
+		}
+		if m != nil {
+			t.Fatalf("Expected nil message, got %v", m)
+		}
+	}
+
+	check("should have headers")
+
+	msg.Header.Set(statusHdr, noMessagesSts)
+	check(ErrMsgNotFound.Error())
+
+	msg.Header.Set(statusHdr, reqTimeoutSts)
+	check("unable to get message")
+
+	msg.Header.Set(descrHdr, "some error text")
+	check("some error text")
+
+	msg.Header.Del(statusHdr)
+	msg.Header.Del(descrHdr)
+	msg.Header.Set("some", "header")
+	check("missing stream")
+
+	msg.Header.Set(JSStream, "test")
+	check("missing sequence")
+
+	msg.Header.Set(JSSequence, "abc")
+	check("invalid sequence")
+
+	msg.Header.Set(JSSequence, "1")
+	check("missing timestamp")
+
+	msg.Header.Set(JSTimeStamp, "aaaaaaaaa bbbbbbbbbbbb cccccccccc ddddddddddd eeeeeeeeee ffffff")
+	check("invalid timestamp")
+
+	msg.Header.Set(JSTimeStamp, "2006-01-02 15:04:05.999999999 +0000 UTC")
+	check("missing subject")
+
+	msg.Header.Set(JSSubject, "foo")
+	r, err := convertDirectGetMsgResponseToMsg("test", msg)
+	if err != nil {
+		t.Fatalf("Error during convert: %v", err)
+	}
+	if r.Subject != "foo" {
+		t.Fatalf("Expected subject to be 'foo', got %q", r.Subject)
+	}
+	if r.Sequence != 1 {
+		t.Fatalf("Expected sequence to be 1, got %v", r.Sequence)
+	}
+	if r.Time.UnixNano() != 0xFC4A4D639917BFF {
+		t.Fatalf("Invalid timestamp: %v", r.Time.UnixNano())
+	}
+	if r.Header.Get("some") != "header" {
+		t.Fatalf("Wrong header: %v", r.Header)
+	}
+}
+
+func TestJetStreamConsumerMemoryStorage(t *testing.T) {
+	opts := natsserver.DefaultTestOptions
+	opts.Port = -1
+	opts.JetStream = true
+	s := natsserver.RunServer(&opts)
+	defer shutdownJSServerAndRemoveStorage(t, s)
+
+	nc, js := jsClient(t, s)
+	defer nc.Close()
+
+	if _, err := js.AddStream(&StreamConfig{Name: "STR", Subjects: []string{"foo"}}); err != nil {
+		t.Fatalf("Error adding stream: %v", err)
+	}
+
+	// Pull ephemeral consumer with memory storage.
+	sub, err := js.PullSubscribe("foo", "", ConsumerMemoryStorage())
+	if err != nil {
+		t.Fatalf("Error on subscribe: %v", err)
+	}
+
+	consInfo, err := sub.ConsumerInfo()
+	if err != nil {
+		t.Fatalf("Error getting consumer info: %v", err)
+	}
+
+	if !consInfo.Config.MemoryStorage {
+		t.Fatalf("Expected memory storage to be %v, got %+v", true, consInfo.Config.MemoryStorage)
+	}
+
+	// Create a sync subscription with an in-memory ephemeral consumer.
+	sub, err = js.SubscribeSync("foo", ConsumerMemoryStorage())
+	if err != nil {
+		t.Fatalf("Error on subscribe: %v", err)
+	}
+
+	consInfo, err = sub.ConsumerInfo()
+	if err != nil {
+		t.Fatalf("Error getting consumer info: %v", err)
+	}
+
+	if !consInfo.Config.MemoryStorage {
+		t.Fatalf("Expected memory storage to be %v, got %+v", true, consInfo.Config.MemoryStorage)
+	}
+
+	// Async subscription with an in-memory ephemeral consumer.
+	cb := func(msg *Msg) {}
+	sub, err = js.Subscribe("foo", cb, ConsumerMemoryStorage())
+	if err != nil {
+		t.Fatalf("Error on subscribe: %v", err)
+	}
+
+	consInfo, err = sub.ConsumerInfo()
+	if err != nil {
+		t.Fatalf("Error getting consumer info: %v", err)
+	}
+
+	if !consInfo.Config.MemoryStorage {
+		t.Fatalf("Expected memory storage to be %v, got %+v", true, consInfo.Config.MemoryStorage)
+	}
+}
+
+func TestJetStreamStreamInfoWithSubjectDetails(t *testing.T) {
+	s := RunBasicJetStreamServer()
+	defer shutdownJSServerAndRemoveStorage(t, s)
+
+	nc, js := jsClient(t, s)
+	defer nc.Close()
+
+	var err error
+
+	_, err = js.AddStream(&StreamConfig{
+		Name:     "TEST",
+		Subjects: []string{"test.*"},
+	})
+	if err != nil {
+		t.Fatalf("Unexpected error: %v", err)
+	}
+
+	// Publish on enough subjects to exercise the pagination
+	payload := make([]byte, 10)
+	for i := 0; i < 100001; i++ {
+		nc.Publish(fmt.Sprintf("test.%d", i), payload)
+	}
+
+	// Check that passing a filter returns the subject details
+	result, err := js.StreamInfo("TEST", &StreamInfoRequest{SubjectsFilter: ">"})
+	if err != nil {
+		t.Fatalf("Unexpected error: %v", err)
+	}
+
+	if len(result.State.Subjects) != 100001 {
+		t.Fatalf("expected 100001 subjects in the stream, but got %d instead", len(result.State.Subjects))
+	}
+
+	// Check that passing no filter does not return any subject details
+	result, err = js.StreamInfo("TEST")
+	if err != nil {
+		t.Fatalf("Unexpected error: %v", err)
+	}
+
+	if len(result.State.Subjects) != 0 {
+		t.Fatalf("expected 0 subjects details from StreamInfo, but got %d instead", len(result.State.Subjects))
+	}
+}
+
+func TestStreamNameBySubject(t *testing.T) {
+	s := RunBasicJetStreamServer()
+	defer shutdownJSServerAndRemoveStorage(t, s)
+
+	nc, js := jsClient(t, s)
+	defer nc.Close()
+
+	var err error
+
+	_, err = js.AddStream(&StreamConfig{
+		Name:     "TEST",
+		Subjects: []string{"test.*"},
+	})
+	if err != nil {
+		t.Fatalf("Unexpected error: %v", err)
+	}
+
+	for _, test := range []struct {
+		name       string
+		streamName string
+		err        error
+	}{
+
+		{name: "valid wildcard lookup", streamName: "test.*", err: nil},
+		{name: "valid explicit lookup", streamName: "test.a", err: nil},
+		{name: "lookup on not existing stream", streamName: "not.existing", err: ErrNoMatchingStream},
+	} {
+
+		stream, err := js.StreamNameBySubject(test.streamName)
+		if err != test.err {
+			t.Fatalf("expected %v, got %v", test.err, err)
+		}
+
+		if stream != "TEST" && err == nil {
+			t.Fatalf("returned stream name should be 'TEST'")
+		}
+	}
+}
